# OPTLAB - Copyright (C) 2018 E. Kuci, C. Geuzaine and P. Duysinx, University of Liege
#
# See the LICENSE.txt file for license information. Please report all
# bugs and problems to the public mailing list <optlab@onelab.info>.

# This file defines the OPTLAB Python API (v1.0).
#
# Do not edit it directly: it is automatically generated by `api/gen.py'.
#
# By design, the OPTLAB Python API is purely functional, and only uses elementary
# types (as well as `numpy' arrays if `numpy' is avaiable). See `demos/api' for
# examples.

from ctypes import *
import signal
import os
import platform
from math import pi

OPTLAB_API_VERSION = "1.0"

signal.signal(signal.SIGINT, signal.SIG_DFL)
libdir = os.path.dirname(os.path.realpath(__file__))
if 'Windows' in platform.system() or 'CYGWIN' in platform.system():
    lib = CDLL(os.path.join(libdir, "optlab-1.0.dll"))
elif platform.system() == 'Darwin':
    lib = CDLL(os.path.join(libdir, "liboptlab.dylib"))
else:
    lib = CDLL(os.path.join(libdir, "liboptlab.so"))

use_numpy = False
try:
    import numpy
    try:
        from weakref import finalize as weakreffinalize
    except:
        from backports.weakref import finalize as weakreffinalize
    use_numpy = True
except:
    pass

def _ostring(s):
    sp = s.value.decode("utf-8")
    lib.optlabFree(s)
    return sp

def _ovectorpair(ptr, size):
    if use_numpy:
        v = numpy.ctypeslib.as_array(ptr, (size//2, 2))
        weakreffinalize(v, lib.optlabFree, ptr)
    else:
        v = list((ptr[i * 2], ptr[i * 2 + 1]) for i in range(size//2))
        lib.optlabFree(ptr)
    return v

def _ovectorint(ptr, size):
    if use_numpy:
        v = numpy.ctypeslib.as_array(ptr, (size, ))
        weakreffinalize(v, lib.optlabFree, ptr)
    else:
        v = list(ptr[i] for i in range(size))
        lib.optlabFree(ptr)
    return v

def _ovectordouble(ptr, size):
    if use_numpy:
        v = numpy.ctypeslib.as_array(ptr, (size, ))
        weakreffinalize(v, lib.optlabFree, ptr)
    else:
        v = list(ptr[i] for i in range(size))
        lib.optlabFree(ptr)
    return v

def _ovectorstring(ptr, size):
    v = list(_ostring(cast(ptr[i], c_char_p)) for i in range(size))
    lib.optlabFree(ptr)
    return v

def _ovectorvectorint(ptr, size, n):
    v = [_ovectorint(pointer(ptr[i].contents), size[i]) for i in range(n.value)]
    lib.optlabFree(size)
    lib.optlabFree(ptr)
    return v

def _ovectorvectordouble(ptr, size, n):
    v = [_ovectordouble(pointer(ptr[i].contents), size[i]) for i in range(n.value)]
    lib.optlabFree(size)
    lib.optlabFree(ptr)
    return v

def _ovectorvectorpair(ptr, size, n):
    v = [_ovectorpair(pointer(ptr[i].contents), size[i]) for i in range(n.value)]
    lib.optlabFree(size)
    lib.optlabFree(ptr)
    return v

def _ivectorint(o):
    if use_numpy:
        return numpy.ascontiguousarray(o, numpy.int32).ctypes, c_size_t(len(o))
    else:
        return (c_int * len(o))(*o), c_size_t(len(o))

def _ivectorvectorint(os):
    n = len(os)
    parrays = [_ivectorint(o) for o in os]
    sizes = (c_size_t * n)(*(a[1] for a in parrays))
    arrays = (POINTER(c_int) * n)(*(cast(a[0], POINTER(c_int)) for a in parrays))
    arrays.ref = [a[0] for a in parrays]
    size = c_size_t(n)
    return arrays, sizes, size

def _ivectorvectordouble(os):
    n = len(os)
    parrays = [_ivectordouble(o) for o in os]
    sizes = (c_size_t * n)(*(a[1] for a in parrays))
    arrays = (POINTER(c_double) * n)(*(cast(a[0], POINTER(c_double)) for a in parrays))
    arrays.ref = [a[0] for a in parrays]
    size = c_size_t(n)
    return arrays, sizes, size

def _ivectordouble(o):
    if use_numpy:
        array = numpy.ascontiguousarray(o, numpy.float64)
        ct = array.ctypes
        ct.array = array
        return  ct, c_size_t(len(o))
    else:
        return (c_double * len(o))(*o), c_size_t(len(o))

def _ivectorpair(o):
    if use_numpy:
        array = numpy.ascontiguousarray(o, numpy.int32)
        ct = array.ctypes
        ct.array = array
        return  ct, c_size_t(len(o) * 2)
    else:
        return ((c_int * 2) * len(o))(*o), c_size_t(len(o) * 2)

def _iargcargv(o):
    return c_int(len(o)), (c_char_p * len(o))(*(s.encode() for s in o))


def initialize(type="none"):
    """
    Initialize SCP.
    """
    ierr = c_int()
    lib.optlabInitialize(
        c_char_p(type.encode()),
        byref(ierr))
    if ierr.value != 0:
        raise ValueError(
            "optlabInitialize returned non-zero error code: ",
            ierr.value)

def run():
    """
    Solve the optimization problem.
    """
    ierr = c_int()
    lib.optlabRun(
        byref(ierr))
    if ierr.value != 0:
        raise ValueError(
            "optlabRun returned non-zero error code: ",
            ierr.value)

def addObjective(func):
    """
    Add the objective function.
    """
    ierr = c_int()
    lib.optlabAddObjective(
        c_void_p(func),
        byref(ierr))
    if ierr.value != 0:
        raise ValueError(
            "optlabAddObjective returned non-zero error code: ",
            ierr.value)

def addConstraint():
    """
    Solve the optimization problem.
    """
    ierr = c_int()
    lib.optlabAddConstraint(
        byref(ierr))
    if ierr.value != 0:
        raise ValueError(
            "optlabAddConstraint returned non-zero error code: ",
            ierr.value)

def addGroupOfConstraints():
    """
    Solve the optimization problem.
    """
    ierr = c_int()
    lib.optlabAddGroupOfConstraints(
        byref(ierr))
    if ierr.value != 0:
        raise ValueError(
            "optlabAddGroupOfConstraints returned non-zero error code: ",
            ierr.value)

def addVariable():
    """
    Solve the optimization problem.
    """
    ierr = c_int()
    lib.optlabAddVariable(
        byref(ierr))
    if ierr.value != 0:
        raise ValueError(
            "optlabAddVariable returned non-zero error code: ",
            ierr.value)

def addGroupOfVariables():
    """
    Solve the optimization problem.
    """
    ierr = c_int()
    lib.optlabAddGroupOfVariables(
        byref(ierr))
    if ierr.value != 0:
        raise ValueError(
            "optlabAddGroupOfVariables returned non-zero error code: ",
            ierr.value)

def computeObjective():
    """
    Compute the objective function.
    """
    ierr = c_int()
    lib.optlabComputeObjective(
        byref(ierr))
    if ierr.value != 0:
        raise ValueError(
            "optlabComputeObjective returned non-zero error code: ",
            ierr.value)

def finalize():
    """
    Finalize Optlab.
    """
    ierr = c_int()
    lib.optlabFinalize(
        byref(ierr))
    if ierr.value != 0:
        raise ValueError(
            "optlabFinalize returned non-zero error code: ",
            ierr.value)


class option:
    """
    Global option handling functions
    """

    @staticmethod
    def setNumber(name, value):
        """
        Set a numerical option to `value'. `name'. Available categories and options
        are listed in the Scp reference manual.
        """
        ierr = c_int()
        lib.optlabOptionSetNumber(
            c_char_p(name.encode()),
            c_double(value),
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabOptionSetNumber returned non-zero error code: ",
                ierr.value)

    @staticmethod
    def getNumber(name):
        """
        Get the `value' of a numerical option.

        Return `value'.
        """
        api_value_ = c_double()
        ierr = c_int()
        lib.optlabOptionGetNumber(
            c_char_p(name.encode()),
            byref(api_value_),
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabOptionGetNumber returned non-zero error code: ",
                ierr.value)
        return api_value_.value

    @staticmethod
    def show():
        """
        Print the numerical options.
        """
        ierr = c_int()
        lib.optlabOptionShow(
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabOptionShow returned non-zero error code: ",
                ierr.value)


class mma:
    """
    MMA optimizer
    """

    @staticmethod
    def initialize(currentPoint, lowerBound, upperBound, restart=False):
        """
        Initialize MMA. `numberOfConstraints' ... FIXME TODO
        """
        api_currentPoint_, api_currentPoint_n_ = _ivectordouble(currentPoint)
        api_lowerBound_, api_lowerBound_n_ = _ivectordouble(lowerBound)
        api_upperBound_, api_upperBound_n_ = _ivectordouble(upperBound)
        ierr = c_int()
        lib.optlabMmaInitialize(
            api_currentPoint_, api_currentPoint_n_,
            api_lowerBound_, api_lowerBound_n_,
            api_upperBound_, api_upperBound_n_,
            c_int(bool(restart)),
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabMmaInitialize returned non-zero error code: ",
                ierr.value)

    @staticmethod
    def getCurrentPoint():
        """
        Get current point.

        Return `currentPoint'.
        """
        api_currentPoint_, api_currentPoint_n_ = POINTER(c_double)(), c_size_t()
        ierr = c_int()
        lib.optlabMmaGetCurrentPoint(
            byref(api_currentPoint_), byref(api_currentPoint_n_),
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabMmaGetCurrentPoint returned non-zero error code: ",
                ierr.value)
        return _ovectordouble(api_currentPoint_, api_currentPoint_n_.value)

    @staticmethod
    def updateCurrentPoint(constraints, gradientOfObjective, gradientOfConstraints, diagHessianOfObjective=[]):
        """
        Update current point (based on first order derivatives and possibly the
        diagonal entries of the objective function hessian).
        """
        api_constraints_, api_constraints_n_ = _ivectordouble(constraints)
        api_gradientOfObjective_, api_gradientOfObjective_n_ = _ivectordouble(gradientOfObjective)
        api_gradientOfConstraints_, api_gradientOfConstraints_n_ = _ivectordouble(gradientOfConstraints)
        api_diagHessianOfObjective_, api_diagHessianOfObjective_n_ = _ivectordouble(diagHessianOfObjective)
        ierr = c_int()
        lib.optlabMmaUpdateCurrentPoint(
            api_constraints_, api_constraints_n_,
            api_gradientOfObjective_, api_gradientOfObjective_n_,
            api_gradientOfConstraints_, api_gradientOfConstraints_n_,
            api_diagHessianOfObjective_, api_diagHessianOfObjective_n_,
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabMmaUpdateCurrentPoint returned non-zero error code: ",
                ierr.value)

    @staticmethod
    def getOuterIteration():
        """
        Get outer iteration.

        Return `iteration'.
        """
        api_iteration_ = c_int()
        ierr = c_int()
        lib.optlabMmaGetOuterIteration(
            byref(api_iteration_),
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabMmaGetOuterIteration returned non-zero error code: ",
                ierr.value)
        return api_iteration_.value

    @staticmethod
    def getDesignChange():
        """
        Get the infinite norm between two successive designs.

        Return `change'.
        """
        api_change_ = c_double()
        ierr = c_int()
        lib.optlabMmaGetDesignChange(
            byref(api_change_),
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabMmaGetDesignChange returned non-zero error code: ",
                ierr.value)
        return api_change_.value

    @staticmethod
    def setMoveLimits(lowerBound, upperBound, move):
        """
        Set the outer move limits.
        """
        api_lowerBound_, api_lowerBound_n_ = _ivectordouble(lowerBound)
        api_upperBound_, api_upperBound_n_ = _ivectordouble(upperBound)
        ierr = c_int()
        lib.optlabMmaSetMoveLimits(
            api_lowerBound_, api_lowerBound_n_,
            api_upperBound_, api_upperBound_n_,
            c_double(move),
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabMmaSetMoveLimits returned non-zero error code: ",
                ierr.value)

    @staticmethod
    def finalize():
        """
        Finalize MMA.
        """
        ierr = c_int()
        lib.optlabMmaFinalize(
            byref(ierr))
        if ierr.value != 0:
            raise ValueError(
                "optlabMmaFinalize returned non-zero error code: ",
                ierr.value)


    class option:
        """
        Global option handling functions
        """

        @staticmethod
        def setNumber(name, value):
            """
            Set a numerical option to `value'. `name'. Available categories and options
            are listed in the Scp reference manual.
            """
            ierr = c_int()
            lib.optlabMmaOptionSetNumber(
                c_char_p(name.encode()),
                c_double(value),
                byref(ierr))
            if ierr.value != 0:
                raise ValueError(
                    "optlabMmaOptionSetNumber returned non-zero error code: ",
                    ierr.value)

        @staticmethod
        def getNumber(name):
            """
            Get the `value' of a numerical option.

            Return `value'.
            """
            api_value_ = c_double()
            ierr = c_int()
            lib.optlabMmaOptionGetNumber(
                c_char_p(name.encode()),
                byref(api_value_),
                byref(ierr))
            if ierr.value != 0:
                raise ValueError(
                    "optlabMmaOptionGetNumber returned non-zero error code: ",
                    ierr.value)
            return api_value_.value

        @staticmethod
        def show():
            """
            Print the numerical options.
            """
            ierr = c_int()
            lib.optlabMmaOptionShow(
                byref(ierr))
            if ierr.value != 0:
                raise ValueError(
                    "optlabMmaOptionShow returned non-zero error code: ",
                    ierr.value)
